#!/usr/bin/env python3
import typing
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Callable, Iterator, List, Optional

from bunkai.base.annotation import Annotations, SpanAnnotation


def func_filter_span(
    spans_wide: typing.List[SpanAnnotation], spans_narrow: typing.List[SpanAnnotation]
) -> typing.List[SpanAnnotation]:
    """Compare spans_wide and spans_narrow. If there is an overlap, use wider one."""
    __filtered = []
    for b_ann in spans_narrow:
        is_skip = False
        b_span = b_ann.get_spans()
        for f_ann in spans_wide:
            if b_span.within(f_ann.get_spans()):
                is_skip = True
            else:
                pass
        if is_skip is False:
            __filtered.append(b_ann)
    return __filtered


def func_filter_previous_rule_same_span(
    spans_current: typing.List[SpanAnnotation],
    spans_previous: typing.List[SpanAnnotation],
) -> typing.List[SpanAnnotation]:
    """If there are conflicting results, use the result of the previous rules."""
    spans_current_map = {(sp.start_index, sp.end_index): sp for sp in spans_current}
    spans_previous_map = {(sp.start_index, sp.end_index): sp for sp in spans_previous}
    common_key = list(set(spans_current_map.keys()) & set(spans_previous_map.keys()))
    filtered = []
    for span_key, sp in spans_current_map.items():
        if span_key in common_key:
            continue
        else:
            filtered.append(sp)
    return filtered + spans_previous


class RuleOrderException(Exception):
    """class if the order of rule is out-of-definition."""


class Annotator(metaclass=ABCMeta):
    def __init__(self, rule_name: str):
        self.rule_name = rule_name

    def add_forward_rule(
        self,
        annotation_this_layer: List[SpanAnnotation],
        spans: Annotations,
        func_filtering: Callable = func_filter_previous_rule_same_span,
    ) -> Annotations:
        filtered = func_filtering(annotation_this_layer, spans.get_final_layer())
        spans.add_annotation_layer(self.rule_name, filtered)
        return spans

    @abstractmethod
    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        raise NotImplementedError()


class AnnotationFilter(Annotator):
    @staticmethod
    def unify_span_annotations(
        span_annotations: List[SpanAnnotation],
    ) -> List[SpanAnnotation]:
        span_anns = {str(ann): ann for ann in span_annotations}
        return list(span_anns.values())


class AnnotatorPipeline(metaclass=ABCMeta):
    def __init__(self, pipeline: List[Annotator]):
        self.pipeline = pipeline
        self.check()

    @abstractmethod
    def check(self) -> bool:
        raise NotImplementedError()

    def __iter__(self):
        return iter(self.pipeline)


class SentenceBoundaryDisambiguator(metaclass=ABCMeta):
    def __init__(self, *, path_model: Optional[Path] = None):
        pass

    @abstractmethod
    def eos(self, text: str) -> Annotations:
        raise NotImplementedError()

    @abstractmethod
    def find_eos(self, text: str) -> List[int]:
        raise NotImplementedError()

    @abstractmethod
    def __call__(self, text: str) -> Iterator[str]:
        raise NotImplementedError()



#!/usr/bin/env python3
from pathlib import Path
from typing import Iterator, List, Optional

from bunkai.algorithm.bunkai_sbd.annotator import (
    BasicRule,
    DotExceptionAnnotator,
    EmojiAnnotator,
    EmotionExpressionAnnotator,
    FaceMarkDetector,
    IndirectQuoteExceptionAnnotator,
    LinebreakAnnotator,
    LinebreakForceAnnotator,
    MorphAnnotatorJanome,
    NumberExceptionAnnotator,
)
from bunkai.algorithm.bunkai_sbd.annotator.constant import LAYER_NAME_FIRST
from bunkai.base.annotation import Annotations, SpanAnnotation
from bunkai.base.annotator import AnnotatorPipeline, RuleOrderException, SentenceBoundaryDisambiguator


class BunkaiPipeline(AnnotatorPipeline):
    def check(self) -> bool:
        """
        Check following rule-order.

        1. MorphAnnotatorJanome must be before ExceptionParticle

        2. LinebreakFilter must be in the last position

        3. Facemark detector should be at first position

        """
        class_name2order = {process.rule_name: __i for __i, process in enumerate(self.pipeline)}
        if class_name2order[BasicRule.__name__] < class_name2order[FaceMarkDetector.__name__]:
            raise RuleOrderException(f"{FaceMarkDetector.__name__} should be at first position.")
        if class_name2order[BasicRule.__name__] < class_name2order[EmotionExpressionAnnotator.__name__]:
            raise RuleOrderException(f"{EmotionExpressionAnnotator.__name__} should be before {BasicRule.__name__}.")
        if class_name2order[MorphAnnotatorJanome.__name__] > class_name2order[IndirectQuoteExceptionAnnotator.__name__]:
            raise RuleOrderException(
                f"{MorphAnnotatorJanome.__name__}  must be" f" before {IndirectQuoteExceptionAnnotator.__name__}"
            )
        if LinebreakAnnotator.__name__ in class_name2order:
            if class_name2order[MorphAnnotatorJanome.__name__] > class_name2order[LinebreakAnnotator.__name__]:
                raise RuleOrderException(
                    f"{MorphAnnotatorJanome.__name__}  must be" f" before {LinebreakAnnotator.__name__}"
                )
            if (
                class_name2order[LinebreakAnnotator.__name__]
                > class_name2order[IndirectQuoteExceptionAnnotator.__name__]
            ):
                raise RuleOrderException(
                    f"{LinebreakAnnotator.__name__}  must be" f" before {IndirectQuoteExceptionAnnotator.__name__}"
                )
        return True


class BunkaiSentenceBoundaryDisambiguation(SentenceBoundaryDisambiguator):
    def __init__(self, *, path_model: Optional[Path] = None):
        morph_annotator = MorphAnnotatorJanome()

        _annotators = [
            FaceMarkDetector(),
            EmotionExpressionAnnotator(),
            EmojiAnnotator(),
            BasicRule(),
            morph_annotator,
            IndirectQuoteExceptionAnnotator(),
            DotExceptionAnnotator(),
            NumberExceptionAnnotator(),
        ]

        if path_model is None:
            _annotators.append(LinebreakForceAnnotator())
        else:
            _idxs = [i for i, ann in enumerate(_annotators) if ann.rule_name == MorphAnnotatorJanome.__name__]
            assert len(_idxs) > 0, f"{MorphAnnotatorJanome.__name__} does not exist in a pipeline"
            _annotators.insert(_idxs[0] + 1, LinebreakAnnotator(path_model=path_model))

        self.pipeline = BunkaiPipeline(_annotators)
        super().__init__()

    def eos(self, text: str) -> Annotations:
        annotations = Annotations()
        annotations.add_annotation_layer(
            LAYER_NAME_FIRST,
            [
                SpanAnnotation(
                    rule_name=LAYER_NAME_FIRST,
                    start_index=len(text) - 1,
                    end_index=len(text),
                    split_string_type=None,
                    split_string_value=None,
                )
            ],
        )
        for rule_obj in self.pipeline:
            rule_obj.annotate(text, annotations)
        return annotations

    def find_eos(self, text: str) -> List[int]:
        annotations = self.eos(text)
        end_index = list(sorted(list(set([s_a.end_index for s_a in annotations.get_final_layer()]))))
        return end_index

    def __call__(self, text: str) -> Iterator[str]:
        annotations = self.eos(text)
        end_index = sorted(list(set([s_a.end_index for s_a in annotations.get_final_layer()])))
        __start_index = 0
        __end_index = 0
        for e_i in end_index:
            part_sentences = text[__start_index:e_i]
            __start_index = e_i
            __end_index = e_i
            yield part_sentences

        if __end_index < len(text):
            part_sentences = text[__end_index:]
            yield part_sentences



#!/usr/bin/env python3
import re

from bunkai.base.annotation import Annotations
from bunkai.base.annotator import Annotator


class ExceptionNumeric(Annotator):
    def __init__(self):
        super().__init__(rule_name=self.__class__.__name__)

    @staticmethod
    def is_exception_numeric(original_text: str, start_index: int, end_index: int) -> bool:
        """
        .の前後が数値であった場合は分割を行わない.

        例: 和室3.5畳 / 1.5リットル以上のペットボトル.
        """
        if original_text[start_index:end_index] != "." and original_text[start_index:end_index] != "．":
            return False
        if re.match(r"\d", original_text[start_index - 1]) and re.match(r"\d", original_text[end_index]):
            return True
        return False

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        __return_span_ann = []
        for __s in spans.get_final_layer():
            if self.is_exception_numeric(original_text, __s.start_index, __s.end_index):
                continue
            else:
                __s.rule_name = self.rule_name
                __return_span_ann.append(__s)

        spans.add_annotation_layer(self.rule_name, __return_span_ann)
        return spans



#!/usr/bin/env python3
import re
from pathlib import Path
from typing import List, Optional

from bunkai.algorithm.bunkai_sbd.annotator import constant
from bunkai.base.annotation import SpanAnnotation
from bunkai.base.annotator import Annotations, Annotator

RE_FACEMARK = re.compile(constant.FACE_EXPRESSION_REGEXP)


class FaceMarkDetector(Annotator):
    def __init__(self, *, path_model: Optional[Path] = None):
        super().__init__(FaceMarkDetector.__name__)

    @staticmethod
    def __find_facemark(text: str) -> List[SpanAnnotation]:
        """"""
        __spans = []
        for match_obj in RE_FACEMARK.finditer(text):
            ann = SpanAnnotation(
                rule_name=FaceMarkDetector.__name__,
                start_index=match_obj.regs[0][0],
                end_index=match_obj.regs[0][1],
                split_string_type="facemark",
                split_string_value=text[match_obj.regs[0][0] : match_obj.regs[0][1]],
            )
            __spans.append(ann)
        return __spans

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        span_ann = self.__find_facemark(original_text)
        spans = self.add_forward_rule(span_ann, spans)
        return spans



#!/usr/bin/env python3

import typing
from pathlib import Path

from bunkai.algorithm.bunkai_sbd.annotator.morph_annotator import MorphAnnotatorJanome
from bunkai.algorithm.lbd.predict import Predictor
from bunkai.base.annotation import SpanAnnotation
from bunkai.base.annotator import Annotations, Annotator
from bunkai.constant import METACHAR_LINE_BREAK


class LinebreakAnnotator(Annotator):
    def __init__(self, *, path_model: Path):
        super().__init__(LinebreakAnnotator.__name__)
        self.linebreak_detector = Predictor(modelpath=path_model)

    @staticmethod
    def generate_sentence_structure(
        annotation_object: Annotations, *, splitter: str = "\n", attribute_name: str = "MorphAnnotatorJanome"
    ) -> typing.List[str]:
        input_tokens: typing.List[str] = [
            span_obj.args["token"].word_surface
            for span_obj in annotation_object.get_annotation_layer(attribute_name)
            if span_obj.args is not None
        ]
        sentence_tokens: typing.List[str] = []
        __sentence: str
        for tokens in input_tokens:
            if tokens == splitter:
                sentence_tokens.append(METACHAR_LINE_BREAK)
            else:
                sentence_tokens.append(tokens)

        return sentence_tokens

    @staticmethod
    def merge_preceding_eos(text: str, spans: typing.List[SpanAnnotation]) -> typing.List[SpanAnnotation]:
        # TODO: Make two annotators?
        current_index: int = 0
        processed_spans = []
        sorted_spans = list(sorted(spans, key=lambda s: s.start_index))
        while True:
            if current_index + 1 >= len(sorted_spans):
                break
            span_ann = sorted_spans[current_index]
            if (
                span_ann.end_index == sorted_spans[current_index + 1].start_index
                and sorted_spans[current_index + 1].rule_name == LinebreakAnnotator.__name__
            ):
                processed_spans.append(
                    SpanAnnotation(
                        rule_name=LinebreakAnnotator.__name__,
                        start_index=span_ann.start_index,
                        end_index=sorted_spans[current_index + 1].end_index,
                        split_string_type="linebreak",
                        split_string_value=text[span_ann.start_index : sorted_spans[current_index + 1].end_index],
                    )
                )
                current_index += 2
            else:
                processed_spans.append(span_ann)
                current_index += 1
        return processed_spans

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        """Tokenize済み結果をデータ加工する。Predictorが求める形式にする."""
        sub_texts = self.generate_sentence_structure(spans)
        # tokenizerを更新する。すでにTokenize済みの結果を利用する。
        # self.linebreak_detector.reset_tokenizer(word_tokenizer_type='pre_tokenize', sentence2tokens=sentence2tokens)
        __result = list(self.linebreak_detector.predict([sub_texts]))

        new_spans = spans.get_final_layer()
        morpheme_sequence = list(spans.get_annotation_layer(MorphAnnotatorJanome.__name__))
        if len(__result) > 0:
            # result: typing.List[TokenIndex] = __result[0]  # type: ignore
            for result in __result:
                for predicted_index in result:
                    char_index_start = morpheme_sequence[predicted_index].start_index
                    char_index_end = morpheme_sequence[predicted_index].end_index
                    ann = SpanAnnotation(
                        rule_name=LinebreakAnnotator.__name__,
                        start_index=char_index_start,
                        end_index=char_index_end,
                        split_string_type="linebreak",
                        split_string_value=original_text[char_index_start:char_index_end],
                    )
                    new_spans.append(ann)
            merged_spans = self.merge_preceding_eos(original_text, new_spans)
            spans.add_annotation_layer(self.rule_name, merged_spans)
        else:
            spans.add_annotation_layer(self.rule_name, new_spans)

        return spans



#!/usr/bin/env python3

import argparse
import json
import re
import typing

try:
    from seqeval.metrics import performance_measure
except ImportError:
    raise Exception("You need to install bunkai with pip install -U bunkai[train]")

from bunkai.constant import METACHAR_LINE_BREAK, METACHAR_SENTENCE_BOUNDARY

REGEXP_SB_WITH_BLANKS_SPAN: str = f"({METACHAR_SENTENCE_BOUNDARY})([\\s{METACHAR_LINE_BREAK}]*)"
RE_SB_WITH_BLANKS_SPAN = re.compile(REGEXP_SB_WITH_BLANKS_SPAN)
RE_SBS_SPAN = re.compile(f"{METACHAR_SENTENCE_BOUNDARY}+")


def get_bo(text: str, lbonly: bool) -> typing.List[str]:
    """Get a list of BIO sequence from input text."""
    assert METACHAR_SENTENCE_BOUNDARY + METACHAR_SENTENCE_BOUNDARY not in text
    assert not text.startswith(METACHAR_SENTENCE_BOUNDARY)
    ret: typing.List[str] = []
    exist_under_bar: bool = False
    for char in text:
        if char == METACHAR_SENTENCE_BOUNDARY:
            if lbonly and exist_under_bar is False:
                continue
            else:
                ret[-1] = "SEP"
                exist_under_bar = False
                continue
        if lbonly:
            if char == METACHAR_LINE_BREAK:
                exist_under_bar = True
                ret.append("O")
        else:
            ret.append("O")
    return ret


def get_score(data: typing.Dict[str, int]) -> typing.Dict[str, float]:
    ret: typing.Dict[str, float] = {}
    try:
        precision = data["TP"] / float(data["TP"] + data["FP"])
    except ZeroDivisionError:
        precision = float("nan")
    recall = data["TP"] / float(data["TP"] + data["FN"])
    ret["f1"] = 2 * precision * recall / (recall + precision)
    ret["precision"] = precision
    ret["recall"] = recall
    return ret


def evaluate(golds: typing.List[str], systems: typing.List[str], lbonly: bool) -> str:
    assert len(golds) == len(systems)
    golds_bo = []
    systems_bo = []
    for gold, system in zip(golds, systems):
        assert gold.replace(METACHAR_SENTENCE_BOUNDARY, "") == system.replace(
            METACHAR_SENTENCE_BOUNDARY, ""
        ), f"{gold}\n{system}"

        system_bo = get_bo(system, lbonly)
        systems_bo.append(system_bo)

        gold_bo = get_bo(gold, lbonly)
        golds_bo.append(gold_bo)
        assert len(gold_bo) == len(system_bo), (
            f"gold={gold} \nsystem={system}\n" f"N(gold)={len(gold_bo)} N(systems)={len(system_bo)}"
        )

    measure: typing.Dict[str, int] = performance_measure(golds_bo, systems_bo)
    score = get_score(measure)
    return f"{json.dumps(measure, sort_keys=True)}\n{json.dumps(score, sort_keys=True)}\n"


def trim(val: str) -> str:
    old: str = val
    new: str = val
    while True:
        new = RE_SBS_SPAN.sub(METACHAR_SENTENCE_BOUNDARY, RE_SB_WITH_BLANKS_SPAN.sub(r"\2\1", old))
        if old == new:
            break
        old = new
    return new


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument("--input", "-i", type=argparse.FileType("r"), required=True)
    oparser.add_argument("--gold", "-g", type=argparse.FileType("r"), required=False)
    oparser.add_argument("--output", "-o", type=argparse.FileType("w"), default="-")
    oparser.add_argument("--lb", action="store_true", help='Checks only Line-break marked with "\u2502"')
    oparser.add_argument("--trim", action="store_true")
    return oparser.parse_args()


def main() -> None:
    opts = get_opts()
    with opts.output as outf:
        if opts.trim:
            for line in opts.input:
                outf.write(trim(line[:-1]))
                outf.write("\n")
        else:
            r = evaluate(
                [l1[:-1] for l1 in opts.gold.readlines()],
                [m[:-1] for m in opts.input.readlines()],
                opts.lb,
            )
            outf.write(r)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import sys
import typing
from pathlib import Path

import numpy as np
import torch
from more_itertools import chunked
from transformers.models.auto.modeling_auto import AutoModelForTokenClassification

import bunkai.constant
from bunkai.algorithm.lbd.corpus import LABEL_OTHER, LABEL_SEP, annotation2spans
from bunkai.algorithm.lbd.custom_tokenizers import JanomeSubwordsTokenizer, JanomeTokenizer
from bunkai.algorithm.lbd.train import BunkaiConfig, MyDataset
from bunkai.base.annotation import Tokens
from bunkai.third.utils_ner import InputExample, get_labels

StringMorphemeInputType = typing.List[typing.List[str]]  # (batch-size * variable-length of sentence * tokens)


class Predictor(object):
    def __init__(self, modelpath: Path) -> None:
        """
        Use JanomeTokenizer by default if the input is Document(String).

        If the input is Morpheme(String), Tokenizers are not called.

        :param subword_tokenizer_type: pre_tokenize, janome, basic
        """
        self.model = AutoModelForTokenClassification.from_pretrained(str(modelpath))
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = self.model.to(self.device)

        self.labels = get_labels(str(modelpath.joinpath("labels.txt")))
        self.label_map: typing.Dict[int, str] = {i: label for i, label in enumerate(self.labels)}

        with modelpath.joinpath("bunkai.json").open() as bcf:
            self.bc = BunkaiConfig.from_json(bcf.read())
        # use janome tokenizer or tokenizer based on a vocab-file.
        self.path_tokenizer_model: str = str(Path(modelpath).joinpath("vocab.txt"))
        self.tokenizer = JanomeSubwordsTokenizer(self.path_tokenizer_model)

        # hotfix
        if self.model.base_model_prefix == "distilbert" and "token_type_ids" in self.tokenizer.model_input_names:
            self.tokenizer.model_input_names.remove("token_type_ids")

    def _split_long_text(
        self, tokens: typing.List[str]
    ) -> typing.Tuple[typing.List[typing.List[str]], typing.List[typing.List[int]]]:
        """
        Split documents(tokens) into sub-documents(tokens).

        This is because Bert has the maximum token-length of the input.
        """
        # tokenized_spans_list is a temporary stack which holds subword-token and underbar.
        # That is because underbar is replaced into UNK if underbar is put into a subword-tokeniser.
        tokenized_spans_list: typing.List[typing.List[str]] = []
        tmp_stack: typing.List[str] = []
        for __t in tokens:
            if __t == bunkai.constant.METACHAR_LINE_BREAK:
                if len(tmp_stack) > 0:
                    tokenized_spans_list.append(tmp_stack)
                tokenized_spans_list.append([bunkai.constant.METACHAR_LINE_BREAK])
                tmp_stack = []
            else:
                # sub-word tokenize
                tmp_stack.append(__t)
        if len(tmp_stack) > 0:
            tokenized_spans_list.append(tmp_stack)

        # run subword-tokeniser
        processed_tokens: typing.List[typing.List[str]] = [[]]
        processed_num_sws: typing.List[typing.List[int]] = [[]]
        current_count: int = 0
        tokenized_span: typing.List[str]
        for tokenized_span in tokenized_spans_list:
            for word in tokenized_span:
                subwords = self.tokenizer.tokenize(word)
                num_subwords = len(subwords)
                current_count += num_subwords
                if current_count >= self.bc.max_seq_length:
                    processed_tokens.append([])
                    processed_num_sws.append([])
                    current_count = num_subwords
                    assert current_count < self.bc.max_seq_length
                processed_tokens[-1].append(word)
                processed_num_sws[-1].append(num_subwords)
        return processed_tokens, processed_num_sws

    def predict(self, documents_morphemes: StringMorphemeInputType) -> typing.List[typing.Set[int]]:
        """
        Run prediction on incoming inputs. Inputs are 2 dims array with [[sentence]].

        :param spans_list: 2 dims list (batch-size * variable-length of sentence) or
        [['ラウンジ', 'も', '気軽', 'に', '利用', 'でき', '、', '申し分', 'ない', 'です', '。', '▁', '']].
        """
        examples = []

        # Note: gave up to separate this process in MyDataset because _split_long_text calls a tokenizer.
        num_subwords = []
        for d_id, spans in enumerate(documents_morphemes):
            sentences_within_length, _num_subwords = self._split_long_text(spans)
            num_subwords += _num_subwords
            for local_s_id, tokenized_spans in enumerate(sentences_within_length):
                examples.append(
                    InputExample(
                        guid=f"{d_id}-{local_s_id}",
                        words=tokenized_spans,
                        labels=[LABEL_OTHER] * len(tokenized_spans),
                        is_document_first=bool(local_s_id == 0),
                    )
                )
        ds = MyDataset(examples, self.labels, self.bc.max_seq_length, self.tokenizer, False)

        kwargs = {
            "input_ids": torch.tensor([f.input_ids for f in ds.features], device=self.device),
            "attention_mask": torch.tensor([f.attention_mask for f in ds.features], device=self.device),
        }
        if self.model.base_model_prefix == "bert":  # hotfix
            kwargs["token_type_ids"] = torch.tensor([f.token_type_ids for f in ds.features], device=self.device)

        predictions = self.model(**kwargs).logits.to("cpu").detach().numpy()

        assert isinstance(
            predictions, np.ndarray
        ), f"Unexpected error. A value type of a model prediction is {type(predictions)}. expect = numpy.ndarray"
        assert len(predictions.shape) == 3, (
            f"Unexpected error. A value tensor of a model prediction is {len(predictions.shape)} tensor. "
            f"expect = 3rd tensor."
        )

        out: typing.List[typing.Set[int]] = []
        word_idx_offset = 0
        for idx, (example, pred) in enumerate(zip(examples, predictions)):
            sw_idx = 0
            if example.is_document_first:
                word_idx_offset = 0  # reset
                out.append(set())
            else:
                word_idx_offset += len(examples[idx - 1].words)

            for word_idx, _ in enumerate(example.words):
                num_sw: int = num_subwords[idx][word_idx]
                while num_sw > 0:
                    sw_idx += 1
                    label_high_prob = int(np.argmax(pred[sw_idx]))
                    if self.label_map[label_high_prob] == LABEL_SEP:
                        out[-1].add(word_idx + word_idx_offset)
                    num_sw -= 1

        return out


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument("--input", "-i", type=argparse.FileType("r"), required=False, default=sys.stdin)
    oparser.add_argument(
        "--output",
        "-o",
        type=argparse.FileType("w"),
        required=False,
        default=sys.stdout,
    )
    oparser.add_argument("--model", "-m", type=Path, required=True)
    oparser.add_argument("--batch", "-b", type=int, default=1, help="Number of documents to feed a batch")
    return oparser.parse_args()


def generate_initial_annotation_obj(
    input_stream: typing.Iterator[str],
) -> typing.Iterator[typing.List[str]]:
    tokenizer = JanomeTokenizer()
    for document in input_stream:
        # document: a text = document
        assert bunkai.constant.METACHAR_SENTENCE_BOUNDARY not in document
        document_spans: Tokens = annotation2spans(document[:-1])
        document_tokens: typing.List[str] = []
        for fragment in document_spans.spans:
            if bunkai.constant.METACHAR_LINE_BREAK in fragment:
                document_tokens.append(fragment)
            else:
                tokens = tokenizer.tokenize(fragment)
                document_tokens += tokens

        assert "".join(document_tokens) == "".join(document_spans.spans)
        yield document_tokens


def main() -> None:

    opts = get_opts()
    pdt = Predictor(opts.model)

    with opts.input as inf, opts.output as outf:
        for one_batch in chunked(generate_initial_annotation_obj(inf), n=opts.batch):
            for did, token_ids_seps in enumerate(pdt.predict(one_batch)):
                for tid, token in enumerate(one_batch[did]):
                    outf.write(token)
                    if tid in token_ids_seps:
                        outf.write(bunkai.constant.METACHAR_SENTENCE_BOUNDARY)
                else:
                    outf.write("\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import random
import re
import sys
import typing
from pathlib import Path

from bunkai.algorithm.lbd.custom_tokenizers import JanomeTokenizer
from bunkai.base.annotation import Tokens
from bunkai.constant import METACHAR_LINE_BREAK, METACHAR_SENTENCE_BOUNDARY

REGEXP_LB_SPAN: str = (
    f"[\\s{METACHAR_SENTENCE_BOUNDARY}{METACHAR_LINE_BREAK}]*"
    + f"{METACHAR_LINE_BREAK}[\\s{METACHAR_LINE_BREAK}{METACHAR_SENTENCE_BOUNDARY}]*"
)
RE_LB_SPAN = re.compile(REGEXP_LB_SPAN)

LABEL_OTHER: str = "O"
LABEL_SEP: str = "LB_SEP"
LABEL_NSEP: str = "LB_NS"
LABELS: typing.List[str] = [LABEL_OTHER, LABEL_SEP, LABEL_NSEP]


def annotation2spans(sentence: str) -> Tokens:
    """
    Cut a sentence into list of sentence.

    Cation: Return value is Tokens, but this is not actual token. Set of text-segments.
    """
    prev: int = 0
    tokens = Tokens()
    for match in RE_LB_SPAN.finditer(sentence):
        myspan = sentence[prev : match.start()].replace(METACHAR_SENTENCE_BOUNDARY, "")
        if len(myspan) > 0:
            tokens.spans.append(myspan)
            tokens.labels.append(LABEL_OTHER)

        myspan = match.group()
        prev = match.end()
        tokens.spans.append(myspan.replace(METACHAR_SENTENCE_BOUNDARY, ""))
        if METACHAR_SENTENCE_BOUNDARY in myspan:
            tokens.labels.append(LABEL_SEP)
        else:
            tokens.labels.append(LABEL_NSEP)

    lastspan = sentence[prev:]
    if len(lastspan) != 0:
        tokens.spans.append(lastspan.replace(METACHAR_SENTENCE_BOUNDARY, ""))
        tokens.labels.append(LABEL_OTHER)

    assert sentence.replace(METACHAR_SENTENCE_BOUNDARY, "") == "".join(tokens.spans)
    return tokens


def _fix_tokens(mytxt: str, ts: typing.List[str]) -> typing.List[str]:
    # restore blanks erased by the tokeniser
    ret: typing.List[str] = []
    index: int = 0
    prev: int = 0
    for _token in ts:
        index += mytxt[index:].index(_token)
        if index != prev:
            ret.append(mytxt[prev:index])
        ret.append(_token)
        index += len(_token)
        prev = index
    lasttk = mytxt[index:]
    if len(lasttk) > 0:
        ret.append(lasttk)
    return ret


def spans2tokens(tokenizer: JanomeTokenizer, tokens: Tokens) -> Tokens:
    assert len(tokens.spans) == len(tokens.labels)
    new_tokens = Tokens()
    for mytxt, label in zip(tokens.spans, tokens.labels):
        if label == LABEL_OTHER:
            ts = tokenizer.tokenize(mytxt)
            ts = _fix_tokens(mytxt, ts)
            new_tokens.spans += ts
            new_tokens.labels += [LABEL_OTHER] * len(ts)
        else:  # Line break
            new_tokens.spans.append(mytxt)
            new_tokens.labels.append(label)

    return new_tokens


def convert(
    inpath: typing.IO,
    tokenizer: JanomeTokenizer,
    remove_trailing_lb: bool = True,
) -> typing.Iterator[Tokens]:
    with inpath as inf:
        for lid, line in enumerate(inf):
            sentence: str = line[:-1]
            if METACHAR_LINE_BREAK not in sentence:
                continue

            tokens = annotation2spans(sentence)
            if remove_trailing_lb and tokens.labels[-1] != LABEL_OTHER:
                tokens.spans = tokens.spans[:-1]
                tokens.labels = tokens.labels[:-1]
            if len(tokens.labels) == 1:
                continue
            new_tokens = spans2tokens(tokenizer, tokens)
            new_tokens.meta["lid"] = lid
            assert len(new_tokens.labels) == len(new_tokens.spans)
            assert sentence.replace(METACHAR_SENTENCE_BOUNDARY, "").startswith("".join(tokens.spans))
            yield new_tokens


def corpus_separate(
    lines: typing.List[str], ratio_train: float, ratio_dev: float
) -> typing.Dict[str, typing.List[str]]:
    rets: typing.Dict[str, typing.List[str]] = {}

    index_train = int(len(lines) * ratio_train)
    seq_train_set = lines[:index_train]
    seq_dev = seq_train_set[: int(len(seq_train_set) * ratio_dev)]
    seq_train = seq_train_set[int(len(seq_train_set) * ratio_dev) :]
    seq_test = lines[index_train:]

    rets["train"] = seq_train
    rets["test"] = seq_test
    rets["dev"] = seq_dev
    return rets


def pseudo_linebreak(text: str, rnd: random.Random, *, ratio_lb: float = 0.5) -> str:
    rets = []
    text = re.sub(r"。+", METACHAR_LINE_BREAK + METACHAR_SENTENCE_BOUNDARY, text)
    for char in text:
        if char == "、" and rnd.random() < ratio_lb:
            rets.append(METACHAR_LINE_BREAK)
        else:
            rets.append(char)
    return "".join(rets)


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument("--input", "-i", type=argparse.FileType("r"), default=sys.stdin)
    oparser.add_argument("--output", "-o", type=Path, required=True)
    oparser.add_argument("--split", action="store_true")
    oparser.add_argument("--pseudo", action="store_true")
    oparser.add_argument("--train", type=float, default=0.8)
    oparser.add_argument("--dev", type=float, default=0.0)
    oparser.add_argument("--seed", type=int, default=12345)
    oparser.add_argument("--base", default="cl-tohoku/bert-base-japanese-whole-word-masking")
    return oparser.parse_args()


def main() -> None:
    opts = get_opts()

    if opts.split:
        lines = opts.input.readlines()
        random.Random(opts.seed).shuffle(lines)
        name2lines = corpus_separate(lines, ratio_train=opts.train, ratio_dev=opts.dev)
        opts.output.mkdir(exist_ok=True, parents=True)

        ppmap = {LABEL_SEP: METACHAR_SENTENCE_BOUNDARY}
        for name, lines in name2lines.items():
            opath = opts.output.joinpath(f"{name}.jsonl")
            otxtpath = opts.output.joinpath(f"{name}.txt")
            with opath.open("w") as f, otxtpath.open("w") as tf:
                for line in lines:
                    f.write(line)
                    ts = Tokens.from_json(line)
                    tf.write(f"{ts.pretty(ppmap)}\n")
    elif opts.pseudo:
        rnd = random.Random(opts.seed)
        with opts.output.open("w") as outf:
            for line in opts.input:
                line = line[:-1].replace(METACHAR_SENTENCE_BOUNDARY, "").replace(METACHAR_LINE_BREAK, "")
                line2 = pseudo_linebreak(line, rnd)
                outf.write(f"{line2}\n")
    else:
        opts.output.parent.mkdir(exist_ok=True, parents=True)
        tokenizer = JanomeTokenizer(normalize_text=False)
        with opts.output.open("w") as f:
            for tokens in convert(inpath=opts.input, tokenizer=tokenizer):
                f.write(f"{tokens.to_json(ensure_ascii=False, sort_keys=True)}\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import dataclasses
import json
import typing
import unicodedata

import bunkai.constant


@dataclasses.dataclass
class DoccanoDataObject:
    annotation_id: int
    text: str
    labels: typing.List[typing.Tuple[int, int, str]]
    meta: typing.Dict[str, typing.Any]


def sbd(dobj: DoccanoDataObject) -> typing.Iterator[str]:
    __start_index = 0
    labels_sequence = sorted(
        [label_tuple for label_tuple in dobj.labels if label_tuple[2] == "SEP" or label_tuple[2].startswith("SEP-")],
        key=lambda t: t[0],
    )
    for label_tuple in labels_sequence:
        yield dobj.text[__start_index : label_tuple[1]]
        __start_index = label_tuple[1]

    if __start_index != len(dobj.text) - 1:
        yield dobj.text[__start_index:]


def operation(data: dict) -> typing.Iterator[str]:
    dobj = DoccanoDataObject(
        annotation_id=data["id"],
        text=data["text"],
        labels=data["labels"],
        meta=data["meta"],
    )
    for text_part in sbd(dobj):
        if len(text_part) > 0:
            yield text_part


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument("--input", "-i", type=argparse.FileType("r"), default="-", required=True)
    oparser.add_argument("--output", "-o", type=argparse.FileType("w"), default="-")
    oparser.add_argument("--nonfkc", action="store_true")
    return oparser.parse_args()


def main() -> None:
    opts = get_opts()
    with opts.input as inf, opts.output as outf:
        for line in inf:
            data = json.loads(line)
            for __i, string in enumerate(operation(data)):
                if __i > 0:
                    outf.write(bunkai.constant.METACHAR_SENTENCE_BOUNDARY)
                if not opts.nonfkc:
                    string = unicodedata.normalize("NFKC", string)
                outf.write(string)
            outf.write("\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from bunkai.algorithm.bunkai_sbd.annotator.basic_annotator import BasicRule
from bunkai.algorithm.bunkai_sbd.annotator.dot_exception_annotator import DotExceptionAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.emoji_annotator import EmojiAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.emotion_expression_annotator import EmotionExpressionAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.facemark_detector import FaceMarkDetector
from bunkai.algorithm.bunkai_sbd.annotator.indirect_quote_exception_annotator import IndirectQuoteExceptionAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.linebreak_annotator_compat import LinebreakAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.linebreak_force_annotator import LinebreakForceAnnotator
from bunkai.algorithm.bunkai_sbd.annotator.morph_annotator import MorphAnnotatorJanome
from bunkai.algorithm.bunkai_sbd.annotator.number_exception_annotator import NumberExceptionAnnotator

__all__ = [
    "BasicRule",
    "EmotionExpressionAnnotator",
    "FaceMarkDetector",
    "IndirectQuoteExceptionAnnotator",
    "MorphAnnotatorJanome",
    "LinebreakAnnotator",
    "LinebreakForceAnnotator",
    "NumberExceptionAnnotator",
    "DotExceptionAnnotator",
    "EmojiAnnotator",
]



#!/usr/bin/env python3
import re

from bunkai.base.annotation import Annotations, SpanAnnotation
from bunkai.base.annotator import Annotator

RE_LBS = re.compile(r"[\n\s]*\n[\n\s]*")


class LinebreakForceAnnotator(Annotator):
    def __init__(self):
        super().__init__(rule_name=self.__class__.__name__)

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        s2regs = {}
        for r_obj in RE_LBS.finditer(original_text):
            s: int = r_obj.regs[0][0]
            s2regs[s] = r_obj.regs[0]

        __return_span_ann = []

        def _add(ro):
            __return_span_ann.append(
                SpanAnnotation(
                    rule_name=self.rule_name,
                    start_index=s,
                    end_index=ro[1],
                    split_string_type="linebreak",
                    split_string_value=original_text[ro[0] : ro[1]],
                )
            )

        for __s in spans.get_final_layer():
            ro = s2regs.get(__s.end_index)
            if ro is None:
                __return_span_ann.append(__s)
            else:
                _add(ro)
                del s2regs[__s.end_index]

        for ro in s2regs.values():
            _add(ro)

        spans.add_annotation_layer(self.rule_name, __return_span_ann)
        return spans



#!/usr/bin/env python3
import re

from bunkai.algorithm.bunkai_sbd.annotator import constant
from bunkai.base.annotation import Annotations
from bunkai.base.annotator import AnnotationFilter

RE_NUMBER_WORD = re.compile(constant.NUMBER_WORD_REGEXP)


class NumberExceptionAnnotator(AnnotationFilter):
    def __init__(self):
        super().__init__(rule_name=self.__class__.__name__)

    @staticmethod
    def is_exception_no(original_text: str, start_index: int, end_index: int) -> bool:
        """
        .の前にNoがあり、かつ後ろが数字で合った場合には分割を行わない.

        例: おすすめ度No.1 / ROOM No.411.
        """
        if original_text[start_index:end_index] != "." and original_text[start_index:end_index] != "．":
            return False

        if (
            RE_NUMBER_WORD.match(original_text[start_index - 2 : start_index])
            and end_index < len(original_text)
            and re.match(r"\d", original_text[end_index])
        ):
            return True
        return False

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        __return_span_ann = []
        for __s in spans.get_final_layer():
            if self.is_exception_no(original_text, __s.start_index, __s.end_index):
                continue
            else:
                __return_span_ann.append(__s)
        spans.add_annotation_layer(self.rule_name, __return_span_ann)
        return spans



#!/usr/bin/env python3
import re

from bunkai.algorithm.bunkai_sbd.annotator import constant
from bunkai.base.annotation import Annotations, SpanAnnotation
from bunkai.base.annotator import Annotator

PARENT_EMOTION = "|".join(constant.EMOTION_CHARACTERS + constant.EMOTION_EXPRESSIONS)
RE_PARENT_EMOTION = re.compile(
    r"[（\(](" + PARENT_EMOTION + r")[\)）]" + "|" + f"[{constant.EMOTION_SYMBOLS}]+[{constant.PUNCTUATIONS}]?"
)


class EmotionExpressionAnnotator(Annotator):
    def __init__(self):
        super().__init__(rule_name=self.__class__.__name__)

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        reg_points = RE_PARENT_EMOTION.finditer(original_text)

        __return = [
            SpanAnnotation(
                rule_name=self.rule_name,
                start_index=r_obj.regs[0][0],
                end_index=r_obj.regs[0][1],
                split_string_type=EmotionExpressionAnnotator.__name__,
                split_string_value=original_text[r_obj.regs[0][0] : r_obj.regs[0][1]],
            )
            for r_obj in reg_points
        ]
        spans = self.add_forward_rule(__return, spans)
        return spans



#!/usr/bin/env python3
import re

from bunkai.base.annotation import Annotations
from bunkai.base.annotator import AnnotationFilter

NumericExpression = re.compile(r"[〇一二三四五六七八九十百千万億兆京\d]+")
MailaddressCharacter = re.compile(r"[a-zA-Z0-9]")


class DotExceptionAnnotator(AnnotationFilter):
    def __init__(self):
        super().__init__(rule_name=self.__class__.__name__)

    @staticmethod
    def is_exception_numeric(original_text: str, index: int) -> bool:
        """
        Ignore dot between numbers.

        Eg: 和室3.5畳 / 1.5リットル以上のペットボトル.
        """
        if index <= 0:
            return False
        if index + 1 >= len(original_text):
            return False
        if original_text[index] != "．" and original_text[index] != ".":
            return False
        if not NumericExpression.match(original_text[index - 1]):
            return False
        if not NumericExpression.match(original_text[index + 1]):
            return False
        return True

    @staticmethod
    def is_exception_mailaddress(original_text: str, index: int) -> bool:
        if index <= 0:
            return False
        if index + 1 >= len(original_text):
            return False
        if original_text[index] != "．" and original_text[index] != ".":
            return False
        if not MailaddressCharacter.match(original_text[index - 1]):
            return False
        if not MailaddressCharacter.match(original_text[index + 1]):
            return False
        return True

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        __return_span_ann = []
        for __s in spans.get_final_layer():
            if self.is_exception_numeric(original_text, __s.start_index):
                continue
            elif self.is_exception_mailaddress(original_text, __s.start_index):
                continue
            __return_span_ann.append(__s)

        spans.add_annotation_layer(self.rule_name, __return_span_ann)
        return spans



#!/usr/bin/env python3
try:
    import numpy  # noqa: F401
    import torch  # noqa: F401
    import transformers  # noqa: F401
except ImportError:
    install_with_lb = False
else:
    install_with_lb = True

if install_with_lb:
    from bunkai.algorithm.bunkai_sbd.annotator.linebreak_annotator import LinebreakAnnotator  # type: ignore
else:
    # void class for compat
    class LinebreakAnnotator:
        def __init__(self, *, path_model):
            raise Exception("You need to install bunkai with pip install -U bunkai[lb]")



#!/usr/bin/env python3
import re
from typing import Dict, List

from bunkai.algorithm.tsunoda_sbd.annotator import constant
from bunkai.base.annotation import SpanAnnotation
from bunkai.base.annotator import Annotations, Annotator


class ExceptionParentheses(Annotator):
    def __init__(self):
        super().__init__(rule_name=self.__class__.__name__)
        self.re_spans_parentheses1 = re.compile(constant.SPANS_PARENTHESES1_REGEXP)
        self.re_spans_parentheses2 = re.compile(constant.SPANS_PARENTHESES2_REGEXP)

    def replace_parentheses_no1(self, original_text: str, split_points: List[SpanAnnotation]) -> List[SpanAnnotation]:
        """
        括弧内に次の文字列があった場合は、括弧及び括弧内の文字列を一文とする.

        例:  ̃ (近日中には冷房に切り替わる予定です。) ̃ 1 時間飲み放題(カクテル各種! ! )はお勧め.
        """

        def unique_obj(input_list: List[SpanAnnotation]) -> List[SpanAnnotation]:
            # filter out same index
            filtered = []
            __added = []
            for s in input_list:
                if s.end_index not in __added:
                    filtered.append(s)
                __added.append(s.end_index)
            return filtered

        spans_parentheses = self.re_spans_parentheses1.finditer(original_text)
        filtered_split_point = []
        skip_end_index = []
        for parentheses_point in spans_parentheses:
            p_start_index = parentheses_point.regs[0][0]
            p_end_index = parentheses_point.regs[0][1]
            for split_candidate in split_points:
                __split_char = original_text[split_candidate.start_index : split_candidate.end_index]
                if p_start_index < split_candidate.start_index and split_candidate.end_index < p_end_index:
                    if __split_char in constant.CHARS_FOR_IGNORE_PARENTHESES1:
                        # 該当の区切り文字候補は破棄。代わりに()のインデックス情報
                        __new_parentheses_point = SpanAnnotation(
                            rule_name=self.rule_name,
                            start_index=p_end_index - 1,
                            end_index=p_end_index,
                            split_string_type="parentheses-sentence",
                            split_string_value=original_text[p_start_index:p_end_index],
                        )
                        filtered_split_point.append(__new_parentheses_point)
                        skip_end_index.append(split_candidate.end_index)
                    else:
                        pass
                else:
                    pass
            else:
                pass
        else:
            pass

        for s in split_points:
            if s.end_index not in skip_end_index:
                filtered_split_point.append(s)

        return unique_obj(filtered_split_point)

    # FIXME: Duplicated span bug
    def replace_parentheses_no2(self, original_text: str, split_points: List[SpanAnnotation]) -> List[SpanAnnotation]:
        """
        括弧内に、括弧の扱い(1)の文字列ではない基本分割文字列が現れた場合は、二回以上登場した際に限り分割点を付与する.

        要するに、括弧内で文境界付与を与えるということ.

        例: ̃(セルフドリンクサービスはすごく良かったです!種類も豊富。).
        """
        spans_parentheses = list(self.re_spans_parentheses2.finditer(original_text))
        filtered_split_point: List[SpanAnnotation] = []

        target_strings_positions: Dict[int, SpanAnnotation] = {}
        for split_candidate in split_points:
            __split_char = original_text[split_candidate.start_index : split_candidate.end_index]
            if __split_char in constant.CHARS_FOR_IGNORE_PARENTHESES2:
                target_strings_positions[split_candidate.end_index] = split_candidate

        # add split points between parentheses and frequency is more than 2 inside parentheses
        for parentheses_point in spans_parentheses:
            p_start_index = parentheses_point.regs[0][0]
            p_end_index = parentheses_point.regs[0][1]
            __split_char_in_parentheses = [
                (end_pos, reg_obj)
                for end_pos, reg_obj in target_strings_positions.items()
                if p_start_index < end_pos < p_end_index
            ]
            if len(__split_char_in_parentheses) >= 2:
                filtered_split_point += [t[1] for t in __split_char_in_parentheses]

        # add split points outside of parentheses
        t_span_parentheses = [
            (parentheses_point.regs[0][0], parentheses_point.regs[0][1]) for parentheses_point in spans_parentheses
        ]

        def is_outside_of_parentheses(t_s):
            if s_point.split_string_type == "parentheses-sentence":
                return True
            elif s_point.start_index <= t_s[0] and s_point.end_index <= t_s[0]:
                return True
            elif s_point.start_index >= t_s[1] and s_point.end_index >= t_s[1]:
                return True
            else:
                return False

        for s_point in split_points:
            _ = [is_outside_of_parentheses(t_s) for t_s in t_span_parentheses]
            if len(_) > 0 and all(_):
                filtered_split_point.append(s_point)

        return filtered_split_point

    def annotate(self, original_text: str, spans: Annotations) -> Annotations:
        __s_no1 = self.replace_parentheses_no1(original_text, spans.get_final_layer())
        if len(re.findall(r"\(.+\)", original_text)) > 0:
            __s_no2 = self.replace_parentheses_no2(original_text, __s_no1)
        else:
            __s_no2 = __s_no1

        spans.add_annotation_layer(self.rule_name, __s_no2)

        return spans



#!/usr/bin/env python3
import argparse
import collections
import dataclasses
import json
import typing
from pathlib import Path

import tqdm
from dataclasses_json import DataClassJsonMixin

import bunkai.constant
from bunkai.algorithm.lbd.corpus import LABEL_NSEP, LABEL_OTHER, LABEL_SEP, annotation2spans


@dataclasses.dataclass
class Statics(DataClassJsonMixin):
    name: str = ""
    num_file: int = 0
    line_break: int = 0
    line_break_without_sb: int = 0
    sentence: int = 0
    sentence_include_line_break: int = 0
    sentence_include_line_break_without_sb: int = 0

    def add(self, st):
        self.num_file += 1
        for f in dataclasses.fields(st):
            self.__setattr__(f.name, self.__getattribute__(f.name) + st.__getattribute__(f.name))


def count(path: Path, outf: typing.IO, show: bool = False) -> Statics:
    st = Statics()
    with path.open() as inf:
        for doc in inf:
            text = doc[:-1]
            st.sentence += text.count(bunkai.constant.METACHAR_SENTENCE_BOUNDARY)
            tokens = annotation2spans(text)
            if len(tokens.labels) == 0:
                continue
            if tokens.labels[-1] != LABEL_OTHER:
                del tokens.spans[-1]
                del tokens.labels[-1]
            cnt: int = 0
            line_break_without_sb: int = 0
            for label in tokens.labels:
                if label == LABEL_SEP:
                    cnt += 1
                elif label == LABEL_NSEP:
                    cnt += 1
                    line_break_without_sb += 1
            if cnt > 0:
                st.sentence_include_line_break += 1
            st.line_break += cnt

            st.line_break_without_sb += line_break_without_sb
            if line_break_without_sb > 0:
                st.sentence_include_line_break_without_sb += 1
                if show:
                    outf.write(f"{path}\t{tokens}\n")
    return st


def count_char(path: Path) -> typing.Iterator[str]:
    with path.open() as inf:
        for doc in inf:
            text = doc[:-1]
            tokens = annotation2spans(text)
            if len(tokens.labels) == 0:
                continue
            if tokens.labels[-1] != LABEL_OTHER:
                del tokens.spans[-1]
                del tokens.labels[-1]
            for tid, label in enumerate(tokens.labels):
                if tid < len(tokens.labels) - 1 and label == LABEL_OTHER:
                    yield tokens.spans[tid][-1]


def get_opts() -> argparse.Namespace:
    oparser = argparse.ArgumentParser()
    oparser.add_argument("--input", "-i", type=Path, required=True)
    oparser.add_argument("--output", "-o", type=argparse.FileType("w"), default="-")
    oparser.add_argument("--show", action="store_true")
    oparser.add_argument("--char", action="store_true")
    return oparser.parse_args()


def main() -> None:
    opts = get_opts()

    targets: typing.List[typing.Tuple[str, Path]] = [("__", opts.input)]
    if opts.input.is_dir():
        targets = [(fi.name[:2], fi) for fi in opts.input.iterdir()]

    if opts.char:
        genre2chars: typing.Dict[str, typing.DefaultDict[str, int]] = {}
        genre2chars["ALL"] = collections.defaultdict(int)
        for (genre, fpath) in tqdm.tqdm(targets, leave=False):
            chars: typing.Optional[typing.DefaultDict[str, int]] = genre2chars.get(genre)
            if chars is None:
                chars = collections.defaultdict(int)
                genre2chars[genre] = chars
            for mychar in count_char(fpath):
                chars[mychar] += 1
                genre2chars["ALL"][mychar] += 1
        with opts.output as outf:
            for genre, chars in sorted(genre2chars.items()):
                outf.write(f"{genre}\t")
                outf.write(json.dumps(chars, ensure_ascii=False, sort_keys=True))
                outf.write("\n")
        return

    st_all = Statics(name="ALL")
    st_detail: typing.DefaultDict[str, Statics] = collections.defaultdict(Statics)
    with opts.output as outf:
        for (genre, fpath) in tqdm.tqdm(targets, leave=False):
            st = count(fpath, outf, opts.show)
            st_all.add(st)
            st_detail[genre].name = genre
            st_detail[genre].add(st)

        if not opts.show:
            if len(targets) != 1:
                for genre, st in sorted(st_detail.items()):
                    outf.write(f"{st.to_json()}\n")
            outf.write(f"{st_all.to_json()}\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import typing

PUNCTUATIONS: str = "。!?.！？．"
EMOTION_SYMBOLS: str = "…\u2605\u2606\u266a"
EMOTION_CHARACTERS: typing.List[str] = list("笑泣汗涙怒嬉爆驚喜悲謎恥焦苦照憂")
EMOTION_EXPRESSIONS: typing.List[str] = "笑い|わら|泣き|怒り|照れ".split("|")
FACE_SYMBOL_PREFIX_SUFFIX: str = "[!-/:-@[-`{-~！-／：-＠［-｀｛-～\u00A1-\u0FFF\u1000-\u2FFF\u4E00艸]"
ALPHABETS_REGEXP: str = "a-zA-Z0-9ａ-ｚＡ-Ｚ０-９"
FACE_SYMBOL1_REGEXP: str = (
    "[" + ALPHABETS_REGEXP + "!-/:-@[-`{-~！-／：-＠［-｀｛-～ａ-ｚＡ-Ｚ０-９\u00A1-\u0FFF\u1000-\u2FFF\u4E00艸]"
)
FACE_SYMBOL2_REGEXP: str = "[!-/:-@[-`{-~！-／：-＠［-｀｛-～\u00A1-\u0FFF\u1000-\u2FFF\u4E00艸]"
FACE_EXPRESSION_REGEXP: str = (
    FACE_SYMBOL_PREFIX_SUFFIX
    + r"*[（\(]"
    + FACE_SYMBOL1_REGEXP
    + "*"
    + FACE_SYMBOL2_REGEXP
    + "+"
    + FACE_SYMBOL1_REGEXP
    + "*"
    + r"[）\)]"
    + FACE_SYMBOL_PREFIX_SUFFIX
    + "*"
)


NUMBER_WORD_REGEXP: str = r"[nNｎＮ][oOｏＯ]"

LAYER_NAME_FIRST = "first"



#!/usr/bin/env python3
import dataclasses
import typing

import emoji
import emojis

from bunkai.base.annotation import Annotations, SpanAnnotation
from bunkai.base.annotator import Annotator

EMOJI_UNICODE_ENGLISH = {emj: emoji.EMOJI_DATA[emj]["en"] for emj in emoji.EMOJI_DATA}


"""This module detects Emoji"""

# You could set any emoji-category that functions as an end-of-sentence
DEFAULT_TARGET_EMOJI_CATEGORY = ("Smileys & Emotion", "Symbols")


@dataclasses.dataclass
class EmojiText(object):
    start_index: int
    end_index: int
    category: typing.List[typing.Optional[str]]

    def get_span(self) -> int:
        return self.end_index - self.start_index

    def check_emoji_category(self, category: typing.Iterable[str]) -> bool:
        """Be True if the given category-name exists else False."""
        if len(set(category) & set(self.category)) > 0:
            return True
        else:
            return False


class EmojiAnnotator(Annotator):
    def __init__(
        self,
        default_target_emoji_category: typing.Tuple[str, ...] = DEFAULT_TARGET_EMOJI_CATEGORY,
    ):
        super().__init__(rule_name=self.__class__.__name__)
        self.default_target_emoji_category = default_target_emoji_category

    @staticmethod
    def get_emoji_info(emoji_character: str) -> typing.Optional[str]:
        """Get emoji info. return a name of a category."""
        info = emojis.db.get_emoji_by_code(emoji_character)
        try:
            if info is None:
                info = emojis.db.get_emoji_by_code(f"{emoji_character}\ufe0f")
                return info.category
            else:
                return info.category
        except AttributeError:
            return None

    def __find_emoji(self, text: str) -> typing.List[EmojiText]:
        """:return: spans of emoji index. [[start-index, end-index]]."""
        emoji_spans = []
        emoji_categories = []
        is_emoji_span: bool = False
        span_emoji_start: int = 0
        for __i, char in enumerate(text):
            if char not in EMOJI_UNICODE_ENGLISH:
                continue

            emoji_categories.append(self.get_emoji_info(char))
            if not is_emoji_span:
                is_emoji_span = True
                span_emoji_start = __i

            if __i + 1 == len(text) or text[__i + 1] not in EMOJI_UNICODE_ENGLISH:
                emoji_spans.append(EmojiText(span_emoji_start, __i + 1, emoji_categories))
                is_emoji_span = False
                emoji_categories = []
        return emoji_spans

    def annotate(self, original_text: str, spans: Annotations, emoji_threshold: int = 1) -> Annotations:
        __emoji_spans = self.__find_emoji(original_text)
        target_emoji = [e_span for e_span in __emoji_spans if e_span.get_span() >= emoji_threshold]
        target_emoji = [e for e in target_emoji if e.check_emoji_category(self.default_target_emoji_category)]

        __return = [
            SpanAnnotation(
                rule_name=self.rule_name,
                start_index=emoji.start_index,
                end_index=emoji.end_index,
                split_string_type=EmojiAnnotator.__name__,
                split_string_value=original_text[emoji.start_index : emoji.end_index],
                args={"emoji": emoji},
            )
            for emoji in target_emoji
        ]
        spans = self.add_forward_rule(__return, spans)
        return spans
